#!/usr/bin/env python

"""A tiny tool used to test the `convert` plugin. It copies a file and appends
a specified text tag.
"""

from __future__ import division, absolute_import, print_function
import sys
import platform


def convert(in_file, out_file, tag):
    """Copy `in_file` to `out_file` and append the string `tag`.
    """
    # On Python 3, encode the tag argument as bytes.
    if not isinstance(tag, bytes):
        tag = tag.encode('utf8')

    # On Windows, use Unicode paths. (The test harness gives them to us
    # as UTF-8 bytes.)
    if platform.system() == 'Windows':
        in_file = in_file.decode('utf8')
        out_file = out_file.decode('utf8')

    with open(out_file, 'wb') as out_f:
        with open(in_file, 'rb') as in_f:
            out_f.write(in_f.read())
        out_f.write(tag)


if __name__ == '__main__':
    convert(sys.argv[1], sys.argv[2], sys.argv[3])
