body {
    font-family: Helvetica, Arial, sans-serif;
}

#header {
    position: fixed;
    left: 0;
    right: 0;
    top: 0;
    height: 36px;

    color: white;

    cursor: default;

    /* shadowy border */
    box-shadow: 0 0 20px #999;
    -webkit-box-shadow: 0 0 20px #999;
    -moz-box-shadow: 0 0 20px #999;

    /* background gradient */
    background: #0e0e0e;
    background: -moz-linear-gradient(top, #6b6b6b 0%, #0e0e0e 100%);
    background: -webkit-linear-gradient(top, #6b6b6b 0%,#0e0e0e 100%);
}
#header h1 {
    font-size: 1.1em;
    font-weight: bold;
    color: white;
    margin: 0.35em;
    float: left;
}

#entities {
    width: 17em;

    position: fixed;
    top: 36px;
    left: 0;
    bottom: 0;
    margin: 0;

    z-index: 1;
    background: #dde4eb;

    /* shadowy border */
    box-shadow: 0 0 20px #666;
    -webkit-box-shadow: 0 0 20px #666;
    -moz-box-shadow: 0 0 20px #666;
}
#queryForm {
    display: block;
    text-align: center;
    margin: 0.25em 0;
}
#query {
    width: 95%;
    font-size: 1em;
}
#entities ul {
    width: 17em;

    position: fixed;
    top: 36px;
    left: 0;
    bottom: 0;
    margin: 2.2em 0 0 0;
    padding: 0;

    overflow-y: auto;
    overflow-x: hidden;
}
#entities ul li {
    list-style: none;
    padding: 4px 8px;
    margin: 0;
    cursor: default;
}
#entities ul li.selected {
    background: #7abcff;
    background: -moz-linear-gradient(top, #7abcff 0%, #60abf8 44%, #4096ee 100%);
    background: -webkit-gradient(linear, left top, left bottom, color-stop(0%,#7abcff), color-stop(44%,#60abf8), color-stop(100%,#4096ee));
    color: white;
}
#entities ul li .playing {
    margin-left: 5px;
    font-size: 0.9em;
}


#main-detail, #extra-detail {
    position: fixed;
    left: 17em;
    margin: 1.0em 0 0 1.5em;
}
#main-detail {
    top: 36px;
    height: 98px;
}
#main-detail .artist, #main-detail .album, #main-detail .title {
    display: block;
}
#main-detail .title {
    font-size: 1.3em;
    font-weight: bold;
}
#main-detail .albumtitle {
    font-style: italic;
}

#extra-detail {
    overflow-x: hidden;
    overflow-y: auto;
    top: 134px;
    bottom: 0;
    right: 0;
}
/*Fix for correctly displaying line breaks in lyrics*/
#extra-detail .lyrics {
    white-space: pre-wrap;
}
#extra-detail dl dt, #extra-detail dl dd {
    list-style: none;
    margin: 0;
    padding: 0;
}
#extra-detail dl dt {
    width: 10em;
    float: left;
    text-align: right;
    font-weight: bold;
    clear: both;
}
#extra-detail dl dd {
    margin-left: 10.5em;
}


#player {
    float: left;
    width: 150px;
    height: 36px;
}
#player .play, #player .pause, #player .disabled {
    -webkit-appearance: none;
    font-size: 1em;
    font-family: Helvetica, Arial, sans-serif;
    background: none;
    border: none;
    color: white;
    padding: 5px;
    margin: 0;
    text-align: center;

    width: 36px;
    height: 36px;
}
#player .disabled {
    color: #666;
}
